;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  splat-behav.scm
;;;  Behaviors: continuous version of splats.  Don't really have 
;;;  discrete-action semantics.
;;;  
;;;  CVS Identifier:
;;;  $Id$
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  class <splat-behavior>
;;;  splat-behaviors are continuous actions with no discrete-action 
;;;  nature.  Following a wall, for instance.  They can be enabled
;;;  and disabled and that's it.  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <splat-behavior> (<object>)
  (mutex         init-keyword: #f init-value: #f)  
  (enable-thunk  init-keyword: enable-thunk: init-value: #f)
  (disable-thunk init-keyword: disable-thunk: init-value: #f)
  (control-thunk init-keyword: control-thunk: init-value: #f)
  (state         init-keyword: #f init-value: 'disabled)
  (thread        init-keyword: #f init-value: #f)
  (break-flag    init-keyword: #f init-value: #f)
  (break-proc    init-keyword: #f init-value: #f))

(define-method initialize ((self <splat-behavior>) . rest)
  (set-mutex! self (make <mutex>)))

(define-method check-for-disable ((self <splat-behavior>))
  (if (break-flag self)
      ((break-proc self))
      #f))

(define-method eval-behavior-thread ((self <splat-behavior>))
  (set-state! self 'enabled)
  (call-with-current-continuation
   (lambda (break)
     (set-break-proc! self break)
     ((control-thunk self))))
  (set-state! self 'disabled)
  ((disable-thunk self))
  (set-break-flag! self #f))

(define-method enable ((self <splat-behavior>))
  (let ((enable (enable-thunk self)))
    (if (procedure? enable)
        ((enable))))
  (set-thread! self (make <thread> thunk: 
                          (lambda () (eval-behavior-thread self))))
  (mark-thread-ready (thread self)))

(define-method disable ((self <splat-behavior>))
  (set-break-flag! self #t))


