;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  splat-trivia.scm
;;;  Functions used by the splat syntaxifier to build a splat object.
;;;  
;;;  CVS Identifier:
;;;  $Id$
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define-method add-initializer! ((self <splat>)
                                 (new-thunk <closure>))
  (with-semaphore (mutex self)
              (set-initializer-list! 
               self 
               (cons new-thunk (initializer-list self)))))

(define-method add-finalizer! ((self <splat>)
                               (new-thunk <closure>))
  (with-semaphore (mutex self)
              (set-finalizer-list! 
               self 
               (cons new-thunk (finalizer-list self)))))

(define-method add-precondition! ((self <splat>) 
                                  (new-thunk <closure>))
  (with-semaphore (mutex self)
              (set-precond-list! self (cons new-thunk (precond-list self)))))

(define-method add-wait-pred! ((self <splat>) 
                               (new-thunk <splat-predicated-action>))
  (with-semaphore (mutex self)
                  (set-wait-pred-list! 
                   self (cons new-thunk (wait-pred-list self)))))

(define-method add-event-handler! ((self <splat>) 
                                   (new-thunk <splat-predicated-action>))
  (with-semaphore (mutex self)
                  (set-event-handler-list! 
                   self (cons new-thunk (event-handler-list self)))))

(define-method add-method! ((self <splat>) 
                            (new-thunk <splat-method>))
  (with-semaphore (mutex self)
                  (set-method-list! self 
                                    (cons new-thunk (method-list self)))))

(define-method remove-child-splat! ((self <splat>) (child <splat>))
  (set-child-splats! self 
                    (do ((list-left (child-splats self) (cdr list-left))
                         (new-list '()))       
                        ((null? list-left) new-list)
                      (if (not (eq? (car list-left) child))
                          (set! new-list (cons (car list-left) new-list))))))

(values #t)