;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  splat.scm
;;;  Class definition and methods for the splat class.
;;;  
;;;  CVS Identifier:
;;;  $Id$
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-class <splat-method> (<object>) 
  (method-tried init-value: #f init-keyword: #f)
  (context      init-value: #f init-keyword: context:)
  (action       init-value: #f init-keyword: action:))

;;; methods for <splat-method>
(define-generic-function execute)

(define-class <splat> (<object>) 
  (name                init-value: 'unnamed init-keyword: name:)
  (precond-list        init-value: '() init-keyword: #f)
  (initializer-list    init-value: '() init-keyword: #f)
  (event-handler-list  init-value: '() init-keyword: #f)
  (wait-pred-list      init-value: '() init-keyword: #f)
  (method-list         init-value: '() init-keyword: #f)
  (finalizer-list      init-value: '() init-keyword: #f)
  (initializers-run    init-value: #f  init-keyword: #f)
  (child-splats        init-value: '() init-keyword: #f)
  (waiting-cond        init-value: #f  init-keyword: #f)
  (mutex               init-value: #f  init-keyword: #f)
  (condition           init-value: #f  init-keyword: #f)
  (thread              init-value: #f  init-keyword: #f) 
  (state               init-value: #f  init-keyword: #f)
  (return-value        init-value: #f  init-keyword: #f)
  (exit-proc           init-value: #f  init-keyword: #f)) 

;; splat methods 
(define-generic-function eval-splat-thread)
;(define-generic-function start-task)
(define-generic-function wait-for-task)
(define-generic-function wait-for-pred)
(define-generic-function finish)
(define-generic-function method-finish)
(define-generic-function add-method!)
(define-generic-function add-event-handler!)
(define-generic-function add-precondition!)
(define-generic-function add-initializer!)
(define-generic-function add-finalizer!)
(define-generic-function add-wait-pred!)
(define-generic-function remove-child-splat!)
(define-generic-function check-completion)
(define-generic-function interruptible-wait)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  <splat-method> methods
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define-method initialize ((self <splat-method>) . other-args)
  (if (not (procedure? (context self)))
      (set-context! self (lambda (splat) #t)))
  (if (not (procedure? (action self)))
      (set-action! self (lambda (splat) #t))))

(define (splat-method? obj)
  (subclass? (object-class obj) <splat-method>))

(define-method execute ((meth <splat-method>) (spl <splat>)) 
  (let ((retval #f))
    (set-method-tried! meth #t)
    (set! retval ((action meth)))
    (if retval (set! retval 'method-succeeded))
    (method-finish spl retval)))
  
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  <splat> methods
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (splat? obj) 
  (subclass? (object-class obj) <splat>))

(define-method initialize ((self <splat>) . other-args)
  (set-state! self 'inactive)
  (set-condition! self (make <condition-variable>))
  (set-mutex! self (make <mutex>)))

;;;  (finish self return-code)
;;;  finish causes the splat to finish, regardless of whether the 
;;;  return-code is true or false.  If it's false, no other method 
;;;  will be tried.  

(define-method finish ((self <splat>) return-code) 
  (for-each (lambda (this-child)
              (finish this-child return-code))
            (child-splats self))
  (with-semaphore (mutex self)
              (set-state! self 'finished)
              (set-return-value! self return-code))
  (condition-broadcast (condition self))

  ;; if this thread is the method thread of self, invoke the 
  ;; exit-continuation 
  (if (eq? (thread self) *current-thread*)
      (if (exit-proc self)
          ((exit-proc self)))
      ;; this thread is not the thread that needs to die.
      ;; If this splat is sleeping, wake it up. 
      (if (waiting-cond self)
          (condition-broadcast (waiting-cond self)))))

(define-method method-finish ((self <splat>) return-code)  
  ;; still want any child splats to finish completely.. we are essentially
  ;; cancelling them.
  (for-each (lambda (this-child)
              (finish this-child return-code))
            (child-splats self))
  
  ;; set self-state to 'method-finished since we want another method
  ;; to be selected.  
  (with-semaphore (mutex self)
;                  (set-state! self 'method-finished) 
                  (set-return-value! self return-code))
;  (condition-broadcast (condition self))
  
  (if (eq? (thread self) *current-thread*)
      (if (exit-proc self)
          ((exit-proc self)))
      (if (waiting-cond self)
          (condition-broadcast (waiting-cond self)))))

(define (start-task (child <splat>) parent)
  (if (eq? (state child) 'running)
      (begin
        (format #t ";;; start-task : error :: task is already active.\n"))
      (begin
        (set-state! child 'running)
        (with-semaphore (mutex child) 
                        (set-thread! 
                         child 
                         (make <thread> 
                               name: (name child)
                               thunk: 
                               (lambda () 
                                 (eval-splat-thread child)))))
        (if (splat? parent)
            (with-semaphore (mutex parent)
                            (set-child-splats! 
                             parent 
                             (append (child-splats parent) (list child)))))
        (mark-thread-ready (thread child))
        (state child))))

(define-method check-completion ((self <splat>))
  (or (eq? (state self) 'finished)
      (eq? (state self) 'method-finished)))

(define-method interruptible-wait 
  ((cv <condition-variable>) (mt <mutex>) self)
  (if (splat? self)
      (set-waiting-cond! self cv))
  (condition-wait cv mt)
  (if (splat? self)
      (begin
        (set-waiting-cond! self #f)
        (if (check-completion self)
            (begin 
              (semaphore-signal mt)
              (if (exit-proc self)
                  ((exit-proc self))))))))

(define-method run-top-level ((target <splat>))
  (start-task target #f)
  (wait-for-task target #f))

(define-method wait-for-task ((target <splat>) parent)
  (let ((rval #f))
    (semaphore-wait (mutex target))
    (do ((st (state target) (state target)))
        ((eqv? st 'finished)
         (begin 
           (if (splat? parent)
               (remove-child-splat! parent target))
           st))
;      (format #t ";; wait-for-task: sleeping for ~s state = ~s~%" 
;              (name target) st)
      (interruptible-wait (condition target) (mutex target) parent))
;     (format #t ";; wait-for-task: woke on ~s~%" 
;              (name target)))
    (set! rval (return-value target))
;    (format #t ";; wait-for-task: done in ~s~%" (name target)) 
    (semaphore-signal (mutex target))
    rval))

(define-method wait-for-pred ((self <splat>) var-list pred)
  (letrec ((condition-mutex (make <mutex>))
           (condition-var (make <condition-variable>))
           (condition-true #f)
           (pa (make <splat-predicated-action>
                     fluents: var-list 
                     predicate: pred 
                     action: (lambda ((self <splat-predicated-action>)) 
                               (set! condition-true #t)
                               (condition-broadcast condition-var)))))
    (enable pa)
    (semaphore-wait condition-mutex)
    (do () 
        (condition-true #t)
      (interruptible-wait condition-var condition-mutex self )
      (if (not (pair? var-list))  ;; special case for wait-forever
          (begin 
            (set! condition-true #t)
            (disable pa))))
    (semaphore-signal condition-mutex)))

(define-method wait-forever ((self <splat>))
  (wait-for-pred self '() (lambda (self) #f)))

(values #t)
